"""
    spider 中间件
"""
from loguru import logger
from palp.spider.spider_base import BaseSpider


class BaseSpiderMiddleware:
    def spider_start(self, spider: BaseSpider) -> None:
        """
        spider 开始时的操作

        :param spider:
        :return:
        """
        pass

    def spider_close(self, spider: BaseSpider) -> None:
        """
        spider 结束的操作

        :param spider:
        :return:
        """
        pass

    def spider_error(self, spider: BaseSpider, exception_type: str, exception: str) -> None:
        """
        spider 出错时的操作

        :param spider:
        :param exception_type: 错误的类型
        :param exception: 错误的详细信息
        :return:
        """

        logger.error(exception)



import importlib
from typing import Union
from loguru import logger


def import_module(cls_list: Union[list, str]):
    """
    根据字符串导入模块

    :param cls_list: 设置里面导入的对象
    :return:
    """
    class_name_list = []

    if isinstance(cls_list, str):
        cls_list = [cls_list]

    for cls_info in cls_list:
        module, class_name = cls_info.rsplit(".", 1)  # 就是分割成 from module import class_name
        logger.debug(f"正在从 {module} 导入模块：{class_name}")
        cls = importlib.import_module(module).__getattribute__(class_name)  # 导入
        instance = cls()  # 这里是实例化
        class_name_list.append(instance)

    return class_name_list



from palp.filter.filter_bloom import BloomFilter
from palp.filter.filter_redis import RequestRedisFilter
from palp.filter.filter_memory import RequestMemoryFilter
from palp.filter.filter_redis_bloom import RequestRedisBloomFilter



from palp.item.item_base import BaseItem
from palp.network.request import Request
from palp.network.response import Response
from palp.spider.spider_base import BaseSpider
from palp.spider import Spider, DistributiveSpider
from palp.pipeline.pipeline_base import BasePipeline
from palp.network.request_method import RequestGet, RequestPost
from palp.conn.conn_redis import Redis, RedisLock, RedisLockNoWait
from palp.middleware.middleware_spider_base import BaseSpiderMiddleware
from palp.middleware.middleware_request_base import BaseRequestMiddleware
from palp.exception.exception_drop import DropItemException, DropRequestException



"""
    处理爬虫 yield 的请求、item 处理
"""
import inspect
import traceback
from loguru import logger
from palp import settings
from threading import Thread
from palp.network.request import Request
from palp.item.item_base import BaseItem
from palp.exception import exception_drop
from palp.tool.short_module import import_module
from palp.sequence.sequence_base import BaseSequence


class Parser(Thread):
    REQUEST_MIDDLEWARE = []

    def __init__(self, q: BaseSequence, q_item: BaseSequence, spider):
        """

        :param q: request 队列
        :param q_item: item 队列
        :param spider: 运行的爬虫
        """
        super().__init__()
        self.queue = q
        self.queue_item = q_item
        self.spider = spider

        self.waiting = False  # parser 执行状态，如果拿到空任务就是在等待
        self.stop = False  # 是否暂停

    @classmethod
    def from_settings(cls):
        """
        导入过滤设置

        :return:
        """
        request_filter_middleware = import_module(
            settings.REQUEST_FILTER_MIDDLEWARE[settings.SPIDER_TYPE][settings.FILTERING_MODE])
        cls.REQUEST_MIDDLEWARE = request_filter_middleware + import_module(settings.REQUEST_MIDDLEWARE)

    def run(self):
        """
        函数执行入口

        :return:
        """
        while not self.stop:
            task = self.queue.get(timeout=5)

            try:
                if task is None:
                    self.waiting = True
                    continue

                # 将 requests 还原
                elif isinstance(task, Request) and task.callback and hasattr(self.spider, str(task.callback)):
                    task.callback = eval(f"self.spider.{task.callback}")

                self.waiting = False
                self.parse_task(task=task)
            except exception_drop.DropRequestException:
                logger.debug(f"丢弃请求：{task}")
            except Exception as e:
                # spider 报错处理
                for middleware in self.spider.SPIDER_MIDDLEWARE:
                    middleware.spider_error(self.spider, e.__class__.__name__, traceback.format_exc())

    def parse_task(self, task, request: Request = None):
        """
        解析处理任务

        :param task: 任务
        :param request: 上一个的请求
        :return:
        """
        # 判断 yield 的类型
        if isinstance(task, Request):
            if request:
                self.add_new_request(task, request)
            else:
                self.run_requests(task)
        elif isinstance(task, BaseItem):
            self.queue_item.put(task)
        else:
            logger.warning(f"捕获到非法 yield：{task}")

    def add_new_request(self, new_request: Request, old_request: Request):
        """
        添加新的请求到队列中

        :param new_request:
        :param old_request:
        :return:
        """
        new_request.callback = new_request.callback.__name__  # 转成字符串，不然无法序列化
        new_request.session = old_request.session  # 续上上一个的 session
        self.queue.put(new_request)

    def run_requests(self, request: Request):
        """
        处理 yield 的 Request

        :param request:
        :return:
        """
        # 请求发起前的处理
        for middleware in self.__class__.REQUEST_MIDDLEWARE:
            middleware.request_in(self.spider, request)

        # 处理请求
        callback = None
        response = None
        failed_times = 0
        while failed_times < settings.REQUEST_RETRY_TIMES:
            try:
                callback, response = request.send()
                break
            except exception_drop.DropRequestException:
                raise
            except Exception as e:
                failed_times += 1

                # 出现错误可直接处理，并返回新的请求，同时阻断当前错误请求传播
                for middleware in self.__class__.REQUEST_MIDDLEWARE:
                    new_request = middleware.request_error(self.spider, request, e.__class__.__name__,
                                                           traceback.format_exc())
                    if new_request is None:
                        pass
                    elif isinstance(new_request, Request):
                        self.add_new_request(new_request=new_request, old_request=request)
                        raise exception_drop.DropRequestException()
                    else:
                        logger.warning("request_error 仅支持 Request 返回值！")

            # 失败达到最大次数直接丢弃
            if failed_times >= settings.REQUEST_RETRY_TIMES:
                for middleware in self.__class__.REQUEST_MIDDLEWARE:
                    middleware.request_failed(self.spider, request)
                return

        # 请求结束的回调（可在此时判断返回值是否合理，不合理可在此重新构造请求，同时阻断当前请求传播）
        for middleware in self.__class__.REQUEST_MIDDLEWARE:
            new_request = middleware.request_close(self.spider, request, response)
            if new_request is None:
                pass
            elif isinstance(new_request, Request):
                self.add_new_request(new_request=new_request, old_request=request)
                raise exception_drop.DropRequestException()
            else:
                logger.warning("request_close 仅支持 Request 返回值！")

        # 继续监控下次yield
        if callback is not None:
            if inspect.isgeneratorfunction(callback):  # 很好用的库，判断是否是 yield 函数
                for task in callback(request, response):
                    self.parse_task(task=task, request=request)
            else:
                callback(request, response)



"""
    配置
        配置优先级 spider > spider setting > palp setting
        三者的配置是覆盖的关系
"""
from pathlib import Path

SPIDER_TYPE = 1  # 爬虫的类型（1 airspider 2 分布式 spider）（非用户设置）
BASE_PATH = Path(__file__).absolute().parent

# REDIS
REDIS_DB = 0
REDIS_HOST = '127.0.0.1'
REDIS_PORT = 6379
REDIS_PWD = None
REDIS_CLUSTER_NODES = None  # 如果有集群的情况下，优先集群
REDIS_MAX_CONNECTIONS = 5  # redis 最大连接数
REDIS_CONFIG = {}  # redis 连接自定义配置项
REDIS_POOL_CONFIG = {}  # redis pool 连接自定义配置项

# 分布式时的 REDIS KEY
REDIS_KEY_MASTER = '{redis_key}:master'  # redis master（分布式时）
REDIS_KEY_LOCK = '{redis_key}:lock'  # redis 锁
REDIS_KEY_STOP = '{redis_key}:stop'  # 停止所有机器运行（分布式时）
REDIS_KEY_QUEUE_REQUEST = '{redis_key}:request'  # request 队列
REDIS_KEY_QUEUE_FILTER_REQUEST = '{redis_key}:filter:request'  # request 过滤队列
REDIS_KEY_QUEUE_FILTER_ITEM = '{redis_key}:filter:item'  # item 过滤队列
REDIS_KEY_HEARTBEAT = '{redis_key}:heartbeat'  # 机器的心跳（hash）
REDIS_KEY_HEARTBEAT_FAILED = '{redis_key}:heartbeat_failed'  # 校验失败的机器

'''请求相关'''
PERSISTENCE_REQUEST_FILTER = False  # 是否持久化请求过滤（分布式时才有效，否则每次结束都会清除）
REQUEST_FILTER = False  # 去重请求，开启了请求时的 filter_repeat 才有用
REQUEST_RETRY_TIMES = 3  # 请求失败重试次数
REQUEST_TIMEOUT = 10  # 请求超时时间，也可以是元组 (connect timeout, read timeout)
RANDOM_USERAGENT = True  # 如果请求头不含 UA 将会设置，但是自己设置了 UA 则不会设置（默认是 computer，指定则开启下面的选项）
RANDOM_USERAGENT_TYPE = 'computer'  # UA 类型：电脑（computer 代表电脑内随便选，后面代表指定浏览器 chrome、opera、firefox、ie、safari）手机：mobile
REQUEST_PROXIES_TUNNEL_URL = None  # 隧道代理 url

# 请求中间件：处理请求中的各种情况
REQUEST_MIDDLEWARE = [
    "palp.middleware.middleware_request_base.BaseRequestMiddleware",
]

# 请求队列
REQUEST_MODE = 1  # 1 为先进先出队列，2 为后进先出队列，3 为优先级队列
REQUEST_QUEUE = {
    1: {
        1: 'palp.sequence.sequence_memory.FIFOSequence',  # 本地：先进先出队列
        2: 'palp.sequence.sequence_memory.LIFOSequence',  # 本地：后进先出队列
        3: 'palp.sequence.sequence_memory.PrioritySequence',  # 本地：优先级队列（通过 request 的 level 指定，默认 level 10，越小越高）
    },
    2: {
        1: 'palp.sequence.sequence_redis.FIFOSequence',  # redis：先进先出队列
        2: 'palp.sequence.sequence_redis.LIFOSequence',  # redis：后进先出队列
        3: 'palp.sequence.sequence_redis.PrioritySequence',  # redis：优先级队列（通过 request 的 level 指定，默认 level 10，越小越高）
    }
}

# 下载中间件：请求前的处理
PIPELINE_ITEM_BUFFER = 0  # 缓存数量，只有当 item 达到一定数量才会入库，0 为不进行缓存
PIPELINE_RETRY_TIMES = 3  # 入库失败重试次数
PIPELINE = [
    "palp.pipeline.pipeline_base.BasePipeline",
]

'''过滤去重'''
BLOOMFILTER_BIT = 6
BLOOMFILTER_HASH_NUMBER = 30
FILTERING_MODE = 2  # 去重方式：1 为 set 集合，2 为 bloom 过滤
ITEM_FILTER = False  # 是否去重 item 默认不去重
PERSISTENCE_ITEM_FILTER = False  # 是否持久化 item 过滤（分布式时才有效，否则每次结束都会清除）

# item 过滤中间件（1、本地，2、云端分布式）
ITEM_FILTER_PIPELINE = {
    1: {
        1: 'palp.pipeline.pipeline_item_filter.ItemMemoryFilterPipeline',  # 本地：set 过滤
        2: 'palp.pipeline.pipeline_item_filter.ItemBloomFilterPipeline',  # 本地：布隆过滤
    },
    2: {
        1: 'palp.pipeline.pipeline_item_filter.ItemRedisFilterPipeline',  # redis：set 过滤
        2: 'palp.pipeline.pipeline_item_filter.ItemRedisBloomFilterPipeline'  # redis：布隆过滤
    }
}

# 请求过滤中间件（1、本地，2、云端分布式）
REQUEST_FILTER_MIDDLEWARE = {
    1: {
        1: 'palp.middleware.middleware_request_filter.RequestMemoryFilterMiddleware',  # 本地：set 过滤
        2: 'palp.middleware.middleware_request_filter.RequestBloomFilterMiddleware',  # 本地：布隆过滤
    },
    2: {
        1: 'palp.middleware.middleware_request_filter.RequestRedisFilterMiddleware',  # redis：set 过滤
        2: 'palp.middleware.middleware_request_filter.RequestRedisBloomFilterMiddleware'  # redis：布隆过滤
    }
}

# spider 中间件
SPIDER_MIDDLEWARE = [
    'palp.middleware.middleware_spider_base.BaseSpiderMiddleware',
]

# 爬虫本身配置
SPIDER_THREAD_COUNT = 16  # spider 默认的线程数量

# 预警：Email
EMAIL_USER = None  # Email 账号
EMAIL_PWD = None  # Email 授权码
EMAIL_RECEIVER = []  # Email 接收者列表

# 预警：DingTalk
DING_TALK_AT_MOBILES = []  # @ 手机号
DING_TALK_AT_USER_IDS = []  # @ user_id
DING_TALK_AT_ALL = False  # @ 所有人
DING_TALK_SECRET = None  # 加签
DING_TALK_ACCESS_TOKEN = None  # webhook 链接内的 access_token

# 日志（详见：https://loguru.readthedocs.io/en/latest/）
LOG_LEVEL = "DEBUG"  # 日志过滤等级
LOG_SHOW = True  # 是否打印日志
LOG_SAVE = False  # 是否保存日志
LOG_PATH = BASE_PATH.joinpath('log')  # 日志存放地址
LOG_ROTATION = '5 MB'  # 日志切割方式，这里是 5m
LOG_RETENTION = '5 days'  # 日志清理时间
LOG_COMPRESSION = 'zip'  # 日志压缩方式（gz、bz2、xz、lzma、tar、tar.gz、tar.bz2、tar.xz、zip）

# 导入 用户的 spider 设置
try:
    from settings import *
except:
    pass



"""
    返回带有解析器功能的 Response
"""

import re
import requests
from typing import Any
from parsel import Selector
from requests.models import Response as RequestResponse


# 响应解析器
class Response(RequestResponse):
    def __init__(self, resp: RequestResponse):
        """

        :param resp: requests 返回的响应
        """
        super().__init__()
        self._resp = resp
        self._resp.encoding = resp.apparent_encoding
        self._selector = Selector(self._resp.text)

        self.__dict__.update(self._resp.__dict__)  # 把 request Response 对象的数据转移过来

    def json(self, **kwargs) -> Any:
        return self._resp.json(**kwargs)

    def xpath(self, query: str, namespaces: str = None, **kwargs):
        """
        实现 xpath 选择器功能

        :param query: 匹配规则
        :param namespaces: 命名空间
        :param kwargs:
        :return:
        """
        return self._selector.xpath(query=query, namespaces=namespaces, **kwargs)

    def css(self, query: str):
        """
        实现 css 选择器功能

        :param query: 匹配规则
        :return:
        """
        return self._selector.css(query=query)

    def re(self, pattern: str, flags=0) -> list:
        """
        常用 flags：
            re.I：忽略字母大小写
            re.A：匹配相应的 ASCII 字符类别

        :param pattern: 匹配规则
        :param flags: 修饰符
        :return:
        """
        return re.findall(pattern, self.text, flags=flags)

    def re_first(self, pattern: str, flags=0) -> str:
        """
        常用 flags：
            re.I：忽略字母大小写
            re.A：匹配相应的 ASCII 字符类别

        :param pattern: 匹配规则
        :param flags: 修饰符
        :return:
        """
        result = self.re(pattern=pattern, flags=flags)
        if result:
            return result[0]

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __getattr__(self, item):
        return self.__dict__.get(item)


# 发起请求，返回响应，含有 session 功能
class ResponseDownloader:

    def __init__(self, keep_session: bool, session: requests.Session):
        """

        :param keep_session: 是否需要保持 session
        :param session: session
        """
        self.keep_session = keep_session
        self.session = session

    def response(self, **kwargs) -> Response:
        if self.keep_session:
            resp = self.session.request(**kwargs)
        else:
            resp = requests.request(**kwargs)

            # 更新请求的 cookie、响应的 cookie
            self.session.cookies.update(kwargs['cookies'])
            self.session.cookies.update(resp.cookies.get_dict())

        return Response(resp)
